import functools
import itertools as it
from typing import Iterable, Tuple, TypeVar, Callable, Any

A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")

Func = TypeVar("Func", bound=Callable[..., Any])


def curry(func: Func) -> Callable[..., Any]:
    """Apply automatic currying to a python function"""

    @functools.wraps(func)
    def curried(*args, **kwargs):
        if func.__code__.co_argcount <= len(args):
            return func(*args, **kwargs)
        return functools.partial(func, *args, **kwargs)

    return curried


def identity(value: A) -> A:
    """The identity function"""
    return value


def apply(func: Func, *args, **kwargs) -> A:
    return func(*args, **kwargs)


@curry
def const(value: A, _) -> A:
    """ Simply returns it's first argument"""
    return value


@curry
def compose(f: Callable[[B], C], g: Callable[[A], B]) -> Callable[[A], C]:
    return lambda x: f(g(x))


@curry
def flip(func: Callable[..., C], a: A, b: B) -> C:
    return func(b, a)


@curry
def partition(
    pred: Callable[[A], bool], iterable: Iterable[A]
) -> Tuple[Iterable[A], Iterable[A]]:
    """Use a predicate to partition entries into false entries and true entries
    
    partition(is_odd, range(10)) --> 0 2 4 6 8   and  1 3 5 7 9
    
    """
    t1, t2 = it.tee(iterable)
    return it.filterfalse(pred, t1), filter(pred, t2)



from __future__ import annotations

from typing import Tuple, TypeVar, Generic, Callable, Union, Iterable, Iterator
from pymonads.applicative import Applicative
from pymonads.monad import Monad
from pymonads.utils import curry, partition


A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")


class _Either(Monad[A]):
    """"""

    @classmethod
    def pure(cls, value):
        return Right(value)

    def is_left(self) -> bool:
        return isinstance(self, Left)

    def is_right(self) -> bool:
        return isinstance(self, Right)


class Left(_Either, Generic[A]):
    """"""

    def fmap(self, func: Callable[[A], B]) -> _Either[B]:
        """"""
        return self

    def amap(self, fab: Applicative[Callable[[A], B]]) -> _Either[B]:
        """"""
        return self

    def flat_map(self, func: Callable[[A], Monad[B]]) -> _Either[B]:
        """"""
        return self


class Right(_Either, Generic[A]):
    """"""

    def fmap(self, func: Callable[[A], B]) -> _Either[B]:
        """"""
        return Right(func(self.value))

    def amap(self, fab: Applicative[Callable[[A], B]]) -> _Either[B]:
        """"""
        return self.fmap(fab.value)

    def flat_map(self, func: Callable[[A], _Either[B]]) -> _Either[B]:
        """"""
        return func(self.value)


Either = Union[Left[A], Right[B]]


@curry
def either(
    left_func: Callable[[A], C], right_func: Callable[[B], C], either_: Either
) -> C:
    val = either_.value
    return left_func(val) if either_.is_left() else right_func(val)


def lefts(eithers: Iterable[Either]) -> Iterator[A]:
    return (either.value for either in eithers if either.is_left())


def rights(eithers: Iterable[Either]) -> Iterator[B]:
    return (either.value for either in eithers if either.is_right())


def from_left(default: A, either_: Either[A, B]) -> A:
    """Return the contents of a Right-value or a default value otherwise."""
    return either_.value if either_.is_left() else default


def from_right(default: B, either_: Either[A, B]) -> B:
    """Return the contents of a Right-value or a default value otherwise."""
    return either_.value if either_.is_right() else default


def partition_eithers(eithers: Iterable[Either]) -> Tuple[Iterable[A], Iterable[B]]:
    return partition(lambda x: x.is_right(), eithers)




"""Functor base class"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, TypeVar, Callable, Any, Type

from pymonads.functor import Functor
from pymonads.utils import identity, curry, compose, const, flip

T = TypeVar('T')

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
D = TypeVar('D')

class Applicative(Functor[A], ABC):
    """An applicative functor"""
    
    @classmethod
    @abstractmethod
    def pure(cls, value: A) -> Applicative[A]:
        """"""

    @abstractmethod
    def amap(self, fab: Applicative[Callable[[A], B]]) -> Applicative[B]:
        """<*>"""

    def sequence_right(self, fb: Applicative[B]) -> Applicative[B]:
        return lift_a_2(flip(const))

    def sequence_left(self, fb: Applicative[B]) -> Applicative[A]:
        return lift_a_2(const)

Ap = Applicative

@curry
def lift_a_1(func: Callable[[A], B], fa: Ap[A]) -> Ap[B]:
    return compose(fa.amap, Applicative.pure)(func)

@curry
def lift_a_2(func: Callable[[A, B], C], fa: Ap[A], fb: Ap[B]) -> Ap[C]:
    """liftA2"""
    return compose(fb.amap, fa.fmap)(func)

@curry
def lift_a_3(func: Callable[[A, B, C], D], fa: Ap[A], fb: Ap[B], fc: Ap[C]) -> Ap[D]:
    return fc.amap(lift_a_2(func, fa, fb))

def pure(applicative_class: Type[Applicative], value: A) -> Applicative[A]:
    """"""
    return applicative_class.pure(value)


"""Functor base class"""

from __future__ import annotations
from dataclasses import dataclass

from functools import partial
from typing import Generic, TypeVar, Callable
from abc import ABC, abstractmethod, abstractproperty

from dataclasses import dataclass

from pymonads.utils import const, compose

T = TypeVar("T")
A = TypeVar("A")
B = TypeVar("B")


@dataclass(frozen=True, repr=False)
class DataClassMixin(Generic[A]):
    """A dataclass mixin.
    
    This is required to get around a mypy error.

    See: https://github.com/python/mypy/issues/5374#issuecomment-568335302
    """

    value: A

class Functor(DataClassMixin[A], ABC):
    """docstring"""

    def __repr__(self):
        return f"{self.__class__.__name__}({self.value})"

    @abstractmethod
    def fmap(self, func: Callable[[A], B]) -> Functor[B]:
        """fmap"""

    def map_replace(self, value: B) -> Functor[B]:
        """The <$ operator"""
        return compose(self.fmap, const)(value)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, TypeVar, Callable, Any

from pymonads.applicative import Applicative
from pymonads.utils import compose

T = TypeVar('T')
A = TypeVar('A')
B = TypeVar('B')

class Monad(Applicative[A], ABC):
    """Base class for Monad implementations"""

    @classmethod
    @abstractmethod
    def pure(cls, value: A) -> Monad[A]:
        """"""

    @abstractmethod
    def flat_map(self, func):
        """Monadic bind or the `>>=` operator."""

    def __or__(self, func: Callable[[A], Monad[B]]) -> Monad[B]:
        """Allows the use of `|` to call flat_map."""
        return self.flat_map(func)

    def __rshift__(self, mb: Monad[B]) -> Monad[B]:
        """Allows the use of `>>` to sequence operations, ignoring the first arg."""
        return mb

    def map(self, func: Callable[[A], B]) -> Monad[B]:
        """Map over values in some monad."""
        def a_to_mb(some_a: A) -> Monad[B]:
            return compose(self.pure, func)(some_a)

        return self.flat_map(a_to_mb)



from setuptools import setup, find_packages

setup(name='pymonads',
      version='0.1.5',
      description='Well typed functional programming abstactions in Python',
      url='http://github.com/amackillop/pymonads',
      author='Austin Mackillop',
      author_email='austin.mackillop@gmail.com',
      license='MIT',
      packages=find_packages(exclude=('tests', 'docs')),
      package_data={
            'pymonads': ['py.typed']
      },
      zip_safe=False)



from __future__ import annotations

from pymonads.applicative import Applicative
from pymonads.monad import Monad
from pymonads.either import Either
from typing import List, TypeVar, Generic, Callable, Union, Iterable, Iterator, Type

from pymonads.utils import curry


from dataclasses import dataclass

A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")


class _Maybe(Monad[A]):
    """"""

    @classmethod
    def pure(cls, value):
        return Just(value)

    def is_just(self) -> bool:
        return isinstance(self, Just)

    def is_nothing(self) -> bool:
        return isinstance(self, Nothing)


class Nothing(_Maybe):
    """"""

    def __init__(self):
        ...

    def __repr__(self):
        return "Nothing"

    @property
    def value(self):
        return None

    def fmap(self, func: Callable[[A], B]) -> _Maybe[B]:
        """"""
        return self

    def amap(self, fab: Applicative[Callable[[A], B]]) -> _Maybe[B]:
        """"""
        return self

    def flat_map(self, func: Callable[[A], Monad[B]]) -> _Maybe[B]:
        """"""
        return self


class Just(_Maybe[A]):
    """"""

    def fmap(self, func: Callable[[A], B]) -> _Maybe[B]:
        return Just(func(self.value))

    def amap(self, fab: Applicative[Callable[[A], B]]) -> _Maybe[B]:
        """"""
        return self.fmap(fab.value)

    def flat_map(self, func: Callable[[A], _Maybe[B]]) -> _Maybe[B]:
        """"""
        return func(self.value)


Maybe = Union[Nothing, Just[A]]


def maybe(default: B, func: Callable[[A], B], maybe_: Maybe[A]) -> B:
    """The maybe function takes a default value, a function, and a Maybe value. 
    
    If the Maybe value is Nothing, the function returns the default value. 
    Otherwise, it applies the function to the value inside the Just and returns the result.
    """
    return func(maybe_.value) if maybe_.is_just() else default


def from_just(maybe_: Maybe[A]) -> A:
    """Extract the element out of a Just or throw an error if its Nothing."""
    if maybe_.is_nothing():
        raise RuntimeError("maybe.from_just called on Nothing")
    return maybe_.value


def from_maybe(default: A, maybe_: Maybe[A]) -> A:
    """The from_maybe function takes a default value and and Maybe value. 
    
    If the Maybe is Nothing, it returns the default value.
    Otherwise, it returns the value contained in the Maybe.
    """
    return maybe_.value if maybe_.is_just() else default


def list_to_maybe(elems: List[A]) -> Maybe[A]:
    """
    The list_to_maybe function returns Nothing on an empty list or 
    Just a where a is the first element of the list.
    """
    return Just(elems[0]) if elems else Nothing()


def maybe_to_list(maybe_: Maybe[A]) -> List[A]:
    """
    The maybeToList function returns an empty list when given Nothing 
    or a singleton list when given Just.
    """
    return [from_just(maybe_)] if maybe_.is_just() else []

def cat_maybes(maybes: Iterable[Maybe[A]]) -> Iterable[A]:
    return (from_just(maybe_) for maybe_ in maybes if maybe_.is_just())

def map_maybe(func: Callable[[A], Maybe[B]], elems: Iterable[A]) -> Iterable[B]:
    for elem in elems:
        val = func(elem)
        if val.is_just():
            yield from_just(val)